from SimulatorFramework import SimUtils
from SimulatorFramework import Frontier
from SimulatorFramework import SimCanvas
from SimulatorFramework import TileAssembly
import RobinsonTiles
import RobinsonAssembly
import wx
import wx.adv


class RobinsonCanvas(SimCanvas.Canvas):
    def __init__(self, parent, idIn, frame, size, app):
        self.drawBitmaps = True
        self.drawSignals = False
        self.drawUpdates = SimCanvas.SIMFW_FULLDISPLAY
        SimCanvas.Canvas.__init__(self, parent, idIn, frame, size, app)
        
    def CreateNewAssembly(self):
        self.sideLen = SimCanvas.CELLSIZE
        self.assembly = RobinsonAssembly.RobinsonAssembly(tileWidths=96, tileHeights=96)
        self.UpdateSizeVars()
        self.RedrawAssembly()
        self.CanvasSizeUpdate((0,0))

    def DoLoadAssembly(self):
        openFileDialog = wx.FileDialog(self.app.window, "Open", "", "", 
                                       "Assembly files (*.txt)|*.txt",
                                       wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)
        openFileDialog.ShowModal()
        fileName = openFileDialog.GetPath()
        openFileDialog.Destroy()
        if (len(fileName) > 0):
            midPt = self.assembly.LoadAssembly(fileName=fileName)
            if midPt == False:
                self.app.OutputMessage("Failed to load assembly from " + fileName)
                return
            self.UpdateSizeVars()
            self.RedrawAssembly(checkAll=True)
            self.CanvasSizeUpdate(midPt)
            self.app.OutputMessage("Loaded assembly from " + fileName)

    def DoSetSeed(self):
        newSeedStr = ""
        msg = "Enter a seed definition"
        dlg = wx.TextEntryDialog(self, msg, caption="Set seed", value=str(newSeedStr))
        if not dlg.ShowModal() == wx.ID_OK:
            dlg.Destroy()
            return

        newSeedStr = dlg.GetValue()
        newSeedStr = str(newSeedStr.strip())
        dlg.Destroy()
        
        midPt = self.assembly.LoadAssembly(seedDefn=newSeedStr)
        if midPt == False:
            self.app.OutputMessage("Failed to set seed assembly to '" + newSeedStr + "'")
            return
        self.RedrawAssembly(checkAll=True)
        self.ScrollToCoords(midPt)
        self.app.OutputMessage("Seed definition set to " + newSeedStr)
        
    def DoSaveAs(self):
        saveFileDialog = wx.FileDialog(self, "Save assembly as", "", "", 
                                       "Assembly files (*.txt)|*.txt",
                                       wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
        if not saveFileDialog.ShowModal() == wx.ID_OK:
            SimUtils.globalApp.ShowErrorDialog("You must provide a file name to save the assembly into.")
            return False
        fileName = saveFileDialog.GetPath()
        saveFileDialog.Destroy()
        if (fileName is None) or len(fileName) < 1:
            return False
        else:
            return self.assembly.SaveAssembly(fileName)

    def DoSetGrowth(self):
        patternString = ""
        patternArray = []
        for pattern in self.assembly.growthPattern:
            if len(patternString) > 0:
                patternString = patternString + ","
            patternString = patternString + pattern
        msg =       "Enter a comma separated list of numbers (0, 1, 2, and 3)\n"
        msg = msg + "to specify the ordering and types of growth tiles selected.\n"
        msg = msg + "To allow all growth tiles at all steps, leave the entire\n"
        msg = msg + "string blank."
        dlg = wx.TextEntryDialog(self, msg, caption="Set growth type patterns", value=str(patternString))
        if dlg.ShowModal() == wx.ID_OK:
            patternArray = []
            newPattern = dlg.GetValue()
            newPattern = str(newPattern.strip())
            if len(newPattern) > 0:
                patternArray = newPattern.split(",")
        dlg.Destroy()
        print("growth pattern input: " + str(self.assembly.growthPattern))
        
        if len(patternArray) == 0:
            self.assembly.growthPattern = []
            self.assembly.growthPatternLoc = 0
        
        # Apply Chaim's pseudocode:
        #=======================================================================
        # For i =  length of the string input sequence, stepping backwards to 1:
        # 
        # Switch on string(i):
        # 
        # If 1 then threeCount:=0; lastnon0State:= last12 := state(i):=  "1";
        # Else if 2, then threeCount: = 0; lastnon0state := last12 := state(i) := "2";
        # Else if 0, then  state(i) := "0" <> lastnon0state;
        # Else if 3, then state(i):=(threeCount++; 
        # Which[
        #         if threeCount == 1, then state(i) =(lastnon0state = "3p" <> last12),
        #         Else if Mod[threeCount, 2] == 1, then state(i) = (lastnon0state = "3d" <> last12),
        #         Else if Mod[threeCount, 2] == 0, then state(i) =(lastnon0state = "3e" <> last12)])
        #      
        # The growth tiles are 
        # "growth_"<>state(i)<>"_"<>state(i+1)
        # for i = 1 to length of string - 1. 
        #=======================================================================

        # If the sequence ends in 0 or 3, append a 1
        #if (patternArray[len(patternArray)-1] == 0) or (patternArray[len(patternArray)-1] == 3):
        #    patternArray.append(1)
        # Actually, just always append a 1 because the output growth pattern will be shortened by 1
        patternArray.append("1")

        states = []
        threeCount = 0
        lastnon0state = 0
        last12 = 0
        for curr in reversed(patternArray):
            print("got " + str(curr))
            if curr == "1":          # If 1 then threeCount:=0; lastnon0State:= last12 := state(i):=  "1";
                threeCount = 0
                lastnon0state = "1"
                last12 = "1"
                state = "1"
            elif curr == "2":        # Else if 2, then threeCount: = 0; lastnon0state := last12 := state(i) := "2";
                threeCount = 0
                lastnon0state = "2"
                last12 = "2"
                state = "2"
            elif curr == "0":        # Else if 0, then  state(i) := "0" <> lastnon0state;
                state = "0" + lastnon0state
            elif curr == "3":        # Else if 3, then state(i):=(threeCount++; 
                                    # Which[
                                    #         if threeCount == 1, then state(i) =(lastnon0state = "3p" <> last12),
                                    #         Else if Mod[threeCount, 2] == 1, then state(i) = (lastnon0state = "3d" <> last12),
                                    #         Else if Mod[threeCount, 2] == 0, then state(i) =(lastnon0state = "3e" <> last12)])
                threeCount = threeCount + 1
                if threeCount == 1:
                    lastnon0state = "3p" + last12
                elif (threeCount % 2 ) == 1:
                    lastnon0state = "3d" + last12
                else:
                    lastnon0state = "3e" + last12
                state = lastnon0state
            states.append(state)
            
        states.reverse()
        print("Modified states: " + str(states))
        self.assembly.growthPattern = []
        for i in range(len(states)-1):
            tile = "growth_" + str(states[i]) + "_" + str(states[i+1])
            self.assembly.growthPattern.append(tile)
        print("Growth pattern: " + str(self.assembly.growthPattern))
        self.assembly.growthPatternLoc = 0
        
    def PrintGrowth(self):
        #print("Frontier: " + str(self.assembly.frontier))
        outStr = "Growth sequence: "
        if len(self.assembly.growthTileLocations) == 0:
            outStr = outStr + " EMPTY"
        else:
            count = 0
            for gLoc in self.assembly.growthTileLocations:
                if count > 0:
                    outStr = outStr + ","
                tile = self.assembly.GetLoc(gLoc)
                if not tile is None:
                    outStr = outStr + tile.type.name
                    #outStr = outStr + "(" + str(gLoc[0]) + "," + str(gLoc[1]) + ")"
                    count = count + 1
        self.app.OutputMessage(outStr)
        
    def PrintMessage(self):
        self.PrintGrowth()
        
    def ShowHelp(self):
        info = wx.adv.AboutDialogInfo()
        info.SetName("Forced Robinson Tiling Simulator")
        info.SetVersion("0.1.0")
        helpStr = "\nList of commands:\n\n"
        helpStr = helpStr + "  l : select tile system definition file to load\n"
        helpStr = helpStr + "  s : save current assembly\n"
        helpStr = helpStr + "  r : reset current tile system to seed\n"
        helpStr = helpStr + "  n : set the seed definition"
        helpStr = helpStr + "  f : simulate one step\n"
        helpStr = helpStr + "  g : begin fast forward simulation\n"
        helpStr = helpStr + "  q : quit fast forward simulation\n"
        helpStr = helpStr + "  p : print list of frontier locations\n"
        helpStr = helpStr + "  t : toggle display of grid lines\n"
        helpStr = helpStr + "  d : toggle display updates of tiles to assembly\n"
        helpStr = helpStr + " +/-: zoom in/out\n"
        helpStr = helpStr + "<left mouse click>:  zoom in on clicked location\n"
        helpStr = helpStr + "<right mouse click>:  zoom out of clicked location\n"
        helpStr = helpStr + "  h : show this help dialog"
        info.SetDescription(helpStr)
        wx.adv.AboutBox(info)

    def SetMouseOverText(self, coords):
        if coords is None:
            self.SendStatusBarUpdate(0, "")
            self.SendStatusBarUpdate(2, "")
            return
        tileTypeName = ""
        tile = self.assembly.GetLoc((coords[0], coords[1]))
        if not tile is None:
            tileTypeName = "Tile type: " + tile.type.name
        else:
            floc = Frontier.FrontierLocation((coords[0], coords[1]),"")
            if floc in self.assembly.frontier:
                tileTypeName = "Frontier location"
                for f in self.assembly.frontier:
                    if f == floc:
                        if f.isGrowthLoc == True:
                            tileTypeName = "Growth location, allows " + str(f.fittingTypeNames)
                        else:
                            
                            tileTypeName = "Frontier location, allows " + str(next(iter(f.fittingTypeNames)))
            
        msg = "Coords: (" + str(int(coords[0])) + "," + str(int(coords[1])) + ")"
        self.SendStatusBarUpdate(0, msg)
        self.SendStatusBarUpdate(2, tileTypeName)
        
    def DoDrawAssemblyUpdate(self, msg):
        if self.drawUpdates == SimCanvas.SIMFW_NODISPLAY:
            return

        if msg == "all":
            self.Refresh()
            return
        elif msg == "allAndCheck":
        #    self.DoInitializeDrawing(dc)
            self.DoRedrawAssembly()

        self._lock.acquire()
               
        dc = wx.ClientDC(self)
            
        self.PrepareDC(dc)
        # we need to clear the dc BEFORE calling PrepareDC
        dc.SetUserScale(self.drawScale, self.drawScale)

        #print("in drawupdate, queue len: " + str(len(self.DrawQueue)))
        if (msg == "new") and self.drawBitmaps == True:
            
            usp = self.CalcUnscrolledPosition(0,0)
            sz = self.GetSize()
            topLeft = self.CoordsFromPixelPos(usp[0], usp[1], True)
            bottomRight = self.CoordsFromPixelPos(usp[0]+sz[0], usp[1]+sz[1], True)

            for e in self.DrawQueue:
                if e[0][0] >= topLeft[0] and e[0][0] <= bottomRight[0] and e[0][1] <= topLeft[1] and e[0][1] >= bottomRight[1]:
                    if e[1] == SimCanvas.SIMFW_DRAWQUEUE_DRAWTILE:
                        loc = self.assembly.GetLoc(e[0])
                        self.DrawAtCoords(dc, e[0][0], e[0][1], loc)
                    elif e[1] == SimCanvas.SIMFW_DRAWQUEUE_DRAWFRONTIER:
                        loc = self.assembly.GetLoc(e[0], includeFrontier=True)
                        self.DrawAtCoords(dc, e[0][0], e[0][1], loc)
                    elif (e[1] == SimCanvas.SIMFW_DRAWQUEUE_REMOVETILE) or (e[1] == SimCanvas.SIMFW_DRAWQUEUE_REMOVEFRONTIER):
                        self.ClearCoord(dc, e[0][0], e[0][1])
                    
            self.DrawQueue = []        
        self._lock.release()

    def OnPaint(self, event):
        self._lock.acquire()

        dc = wx.PaintDC(self)

        bg = wx.Brush("WHITE")#(self.GetBackgroundColour())
        dc.SetBackground(bg)
        dc.Clear()

        # use PrepareDC to set position correctly
        self.PrepareDC(dc)
        # we need to clear the dc BEFORE calling PrepareDC
        
        clipped = dc.GetClippingBox()
        # On OS X, PrepareDC doesn't seem to account for scrolling as it does on Windows
        if "wxOSX" in wx.PlatformInfo:
            xView,yView = self.GetViewStart()
            xDelta,yDelta = self.GetScrollPixelsPerUnit()
            x = clipped[0] + xView * xDelta
            y = clipped[1] + yView * yDelta
        else:
            x = clipped[0]
            y = clipped[1]
        
        topLeft = self.CoordsFromPixelPos(x, y)
        bottomRight = self.CoordsFromPixelPos(x+clipped[2], y+clipped[3])

        dc.SetUserScale(self.drawScale, self.drawScale)
        
        if self.drawGrid == True:
            self.DrawGrid(dc)
        
        if self.drawBitmaps == True:
            #for i in range(self.assembly.minX, self.assembly.maxX+1):
            #    for j in range(self.assembly.minY, self.assembly.maxY+1):
            for i in range(topLeft[0], bottomRight[0]+1):
                for j in range(bottomRight[1], topLeft[1]+1):
                    loc = self.assembly.GetLoc((i,j), includeFrontier=True)
                    if loc is not None:
                        if loc == TileAssembly.FRONTIER_LOC or loc == TileAssembly.FRONTIER_LOC_OFF:
                            self.DrawFrontierLocation(dc, (i,j), loc)
                        elif self.NameMatchesPatterns(loc.type.name) == True:
                            self.DrawAtCoords(dc, i, j, loc)

        #if self.drawSignals == True:
        if self.drawSignals == True:
            font = dc.GetFont()
            font.SetPointSize(12*(1/self.drawScale))
            dc.SetFont(font)
            dc.SetBackgroundMode(wx.SOLID) 
            dc.SetTextBackground(wx.TheColourDatabase.Find("WHITE"))
            dc.SetTextForeground(wx.TheColourDatabase.Find("GREEN"))

            pen = dc.GetPen()
            
            dc.SetPen(wx.Pen("GREEN", (1/self.drawScale)*5))
            cs = self.sideLen*(1/self.drawScale)/5
            c0 = self.GetPixelsForCoords(self.assembly.growthTileLocations[0][0], self.assembly.growthTileLocations[0][1])
            dc.DrawRectangle(c0[0]-(cs/2), c0[1]-(cs/2), cs, cs)

            if len(self.assembly.growthTileLocations) > 1:
                for p in range(0, len(self.assembly.growthTileLocations)-1):
                    c1 = self.GetPixelsForCoords(self.assembly.growthTileLocations[p][0], self.assembly.growthTileLocations[p][1])
                    c2 = self.GetPixelsForCoords(self.assembly.growthTileLocations[p+1][0], self.assembly.growthTileLocations[p+1][1])
                    #print("drawing line (" + str(c1[0]) + "," + str(c1[1]) + ") to (" + str(c2[0]) + "," + str(c2[1]) + ")")
                    #MJPTODO: Add arrowheads to lines so direction is clear
                    dc.SetPen(wx.Pen("RED", (1/self.drawScale)*5))
                    
                    temp = c1
                    c1 = c2
                    c2 = temp
                    dc.DrawLine(c1[0], c1[1], c2[0], c2[1])
                    
                    # Draw directional 'arrow' lines
                    frac = 0.1
                    #midpoint = (int((c1[0]+c2[0])/2), int((c1[1]+c2[1])/2))
                    midpoint = (c1[0], c1[1])
                    slope = (c1[0]-c2[0], c1[1]-c2[1])
                    midpoint = (midpoint[0]-slope[0]*frac, midpoint[1]-slope[1]*frac)
                    invSlope = ((self.sideLen-abs(slope[0]))/self.sideLen, (self.sideLen-abs(slope[1]))/self.sideLen)
        
                    linepoint = (int(midpoint[0]-slope[0]*frac), int(midpoint[1]-slope[1]*frac))
        
                    mult = 10
                    if (slope[0] == 0) or (slope[1] == 0):  # Horizontal
                        pt1 = (int(linepoint[0]+invSlope[0]*7), int(linepoint[1]-invSlope[1]*7))
                        pt2 = (int(linepoint[0]-invSlope[0]*7), int(linepoint[1]+invSlope[1]*7))
                    elif ((slope[0] < 0) and (slope[1] < 0)) or ((slope[0] > 0) and (slope[1] > 0)):
                        pt1 = (int(linepoint[0]+invSlope[0]*mult), int(linepoint[1]-invSlope[1]*mult))
                        pt2 = (int(linepoint[0]-invSlope[0]*mult), int(linepoint[1]+invSlope[1]*mult))
                    else:
                        pt1 = (int(linepoint[0]+invSlope[0]*mult), int(linepoint[1]+invSlope[1]*mult))
                        pt2 = (int(linepoint[0]-invSlope[0]*mult), int(linepoint[1]-invSlope[1]*mult))
                    
                    dc.DrawLine(midpoint[0], midpoint[1], pt1[0], pt1[1])
                    dc.DrawLine(midpoint[0], midpoint[1], pt2[0], pt2[1])
                                        
                    dc.SetPen(wx.Pen("GREEN", (1/self.drawScale)*5))
                    cs = self.sideLen*(1/self.drawScale)/5
                    dc.DrawRectangle(c1[0]-(cs/2), c1[1]-(cs/2), cs, cs)
                    
                    loc = self.assembly.GetLoc((self.assembly.growthTileLocations[p+1][0], self.assembly.growthTileLocations[p+1][1]))
                    if not loc is None:
                        typeName = loc.type.name[7:]
                        dc.DrawText(typeName, c1[0], c1[1])
            dc.SetPen(pen)

        self._lock.release()
        self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles))
        return

    def OnChar(self, evt):
        keycode = evt.GetKeyCode()
        if keycode == ord("n") or keycode == ord("N"):
            self.DoSetSeed()
        elif keycode == ord("a") or keycode == ord("A"):
            self.DoLoadAssembly()
        else:
            SimCanvas.Canvas.OnChar(self, evt)